import pandas as pd
import numpy as np
import datetime
import os
import re

import utils


class ExploProvider:

    def __init__(self):
        self.datasets_root_directory = ""
        self.datasets = {}

        pass

    def explore_root_directory(self, root_directory: str) -> list:
        """Get the names of the datasets directories.

        Parameters
        ----------
        root_directory: str
            Full path of the directory containing the datasets directories.

        Returns
        -------
        list
            List of dataset directories name (without full path)

        """
        directories = []

        # Find all directories in datasets root directory (not recursive)
        for element in os.listdir(root_directory):
            if os.path.isdir(root_directory + "/" + element):
                directories.append(element)

        # Keep only datasets directories (ignore others like pump_calibration, conduct_calib, old, etc.)
        regex = re.compile(r'[0-9]{8}_.*')
        dataset_directories = list(filter(regex.search, directories))

        # Sort list in alphabetical order (in this case, by ascending date)
        dataset_directories.sort()

        self.datasets_root_directory = root_directory
        self.datasets_dirs = dataset_directories
        for directory in dataset_directories:
            dataset = Dataset(root_directory, directory)
            self.datasets[directory] = dataset

        return dataset_directories


class Dataset:

    def __init__(self, root_directory: str, directory_name: str):
        self.root_directory = root_directory
        self.directory_name = directory_name
        self.full_directory_name = root_directory + "/" + directory_name

        # Get dataset name
        self.dataset_text = directory_name[-9:]
        self.last_data_datetime = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)

        self.instlogs = {}

        self.instruments_dict = self.explore_dataset()

    def explore_dataset(self) -> None:
        for filename in os.listdir(self.full_directory_name):
            try:
                inst_and_type = re.search("^" + self.directory_name + '_(.+?).log$', filename).group(1)
            except AttributeError:
                # The found file does not match normal instrument's log file pattern
                print("File [" + filename + "] does not appear to be a valid CFA log file")
                continue

            instrument_name = inst_and_type.split("_")[0]

            if len(inst_and_type.split("_")) == 2:
                log_type = inst_and_type.split("_")[1]
                if log_type == "instant":
                    if instrument_name == "ICBKCTRL":
                        instrument_log = IceblockInstantLog(self.full_directory_name, filename, instrument_name)
                    else:
                        instrument_log = InstrumentInstantLog(self.full_directory_name, filename, instrument_name)
                elif log_type == "periodic":
                    instrument_log = InstrumentPeriodicLog(self.full_directory_name, filename, instrument_name)
                    self.last_data_datetime = max(self.last_data_datetime,
                                                  instrument_log.df["datetime"].max())
                else:
                    raise ValueError("Unknown log type: [" + log_type + "]")
                self.instlogs[inst_and_type] = instrument_log


class InstrumentLog:

    def __init__(self, full_directory_name: str, filename: str, instrument_name: str):
        self.full_directory_name = full_directory_name
        self.filename = filename
        self.full_file_name = full_directory_name + "/" + filename
        self.instrument_name = instrument_name

        self.df = self.__get_df__()

    def get_variables(self):
        raise NotImplementedError("Subclasses should implement this.")

    def get_timeseries(self, variable: str) -> pd.DataFrame:
        raise NotImplementedError("Subclasses should implement this.")

    def __get_df__(self) -> pd.DataFrame:
        raise NotImplementedError("Subclasses should implement this.")


class InstrumentInstantLog(InstrumentLog):

    def __init__(self, full_directory_name: str, filename: str, instrument_name: str):
        InstrumentLog.__init__(self, full_directory_name, filename, instrument_name)

    def __get_df__(self) -> pd.DataFrame:
        df = pd.read_csv(self.full_file_name, sep=",", parse_dates=["datetime"])
        df["datetime"] = df["datetime"].dt.tz_localize('UTC')
        return df

    def get_variables(self):
        return self.df.name.unique()

    def get_timeseries(self, variable: str) -> pd.DataFrame:
        timeseries_df = self.df[self.df["name"] == variable]
        timeseries_df = timeseries_df.drop(columns=['name'])

        try:
            timeseries_df["value"] = timeseries_df["value"].astype(float)
        except ValueError:
            timeseries_df["value_int"] = timeseries_df["value"].astype("category").cat.codes
        return timeseries_df


class IceblockInstantLog(InstrumentLog):

    def __init__(self, full_directory_name: str, filename: str, instrument_name: str):
        InstrumentLog.__init__(self, full_directory_name, filename, instrument_name)

    def __get_df__(self) -> pd.DataFrame:
        df = pd.read_csv(self.full_file_name, sep=",", parse_dates=["datetime"])
        df["datetime"] = df["datetime"].dt.tz_localize('UTC')
        return df

    def get_variables(self):
        return ["melting"]

    def get_timeseries(self, variable: str) -> pd.DataFrame:
        if variable == "melting":
            timeseries_df = self.__get_melting_timeseries__()
        else:
            raise ValueError("Variable name [" + variable + "] not yet managed.")

        return timeseries_df

    def __get_melting_timeseries__(self) -> pd.DataFrame:
        # Get the mapping between iceblock id and iceblock name (assuming that the last name's modification is the
        # good one.
        mapping_df = self.df[["datetime", "id", "name"]].copy()
        mapping_df = mapping_df.groupby("id")["id", "name"].tail(1)
        mapping_df = mapping_df.append({"id": 0, "name": "None"}, ignore_index=True)
        mapping_df = mapping_df.set_index("id")
        mapping_dict = mapping_df["name"].to_dict()

        # Get the datetime of the beginning of each iceblock's melting
        melting_df = self.df[["datetime", "id", "status"]].copy()
        start_df = melting_df[melting_df["status"] == "Melting"].groupby("id")["datetime", "id"].head(1)

        # Get the end of the last iceblock's melting, and set that after that the current melting block is 0/None.
        end_df = melting_df[melting_df["status"] == "Done"].groupby("id").head(1)
        melting_df = start_df.append({"datetime": end_df.iloc[-1]["datetime"], "id": 0},
                                     ignore_index=True)

        # Get the value (iceblocks name) and value_int (coded value, iceblock id in this case).
        melting_df.rename(columns={"id": 'value_int'}, inplace=True)
        melting_df["value"] = melting_df["value_int"].map(mapping_dict)

        return melting_df

class InstrumentPeriodicLog(InstrumentLog):

    def __init__(self, full_directory_name: str, filename: str, instrument_name: str):
        InstrumentLog.__init__(self, full_directory_name, filename, instrument_name)

    def __get_df__(self) -> pd.DataFrame:
        df = pd.read_csv(self.full_file_name, sep="\t", parse_dates=["datetime"])
        df["datetime"] = df["datetime"].dt.tz_localize('UTC')
        return df

    def get_variables(self):
        all_cols = list(self.df)
        variable_cols = [colname for colname in all_cols if colname != "datetime"]
        return variable_cols

    def get_timeseries(self, variable: str) -> pd.DataFrame:
        timeseries_df = self.df[["datetime", variable]]
        timeseries_df.rename(columns={variable: 'value'}, inplace=True)
        return timeseries_df
