import datetime
import pyqtgraph as pg
from PyQt5.QtWidgets import *
from PyQt5.QtGui import QColor
from PyQt5.QtCore import *
import pandas as pd
from pandas.api.types import is_numeric_dtype
import numpy as np
from pyqtgraph.GraphicsScene.mouseEvents import MouseClickEvent

import utils
from config import Config
from dataprovider.exploprovider import ExploProvider
from gui.uimainwindow import Ui_MainWindow


class ExploUim:

    # "Variables" tableWidget columns identifiers.
    INSTRUMENT_COL = 0
    VARIABLE_COL = 1
    COLOR_COL = 2
    OFFSET_COL = 3
    MULT_COL = 4
    TIMESHIFT_COL = 5
    VISIBLE_COL = 6
    XORIG_COL = 7
    YORIG_COL = 8

    DEFAULT_COLORS = [QColor(255, 255, 255),  # White
                      QColor(228, 26, 28),  # Red
                      QColor(55, 126, 184),  # Blue
                      QColor(77, 175, 74),  # Green
                      QColor(152, 78, 163),  # Violet
                      QColor(255, 127, 0),  # Orange
                      QColor(255, 255, 51),  # Yellow
                      QColor(166, 86, 40),  # Brown
                      QColor(247, 129, 191)]  # Pink

    def __init__(self, explo_prvd: ExploProvider, main_ui: Ui_MainWindow, config: Config):
        self.main_ui = main_ui
        self.explo_prvd = explo_prvd
        self.config = config

        self.current_dataset = None

        # The "var_id" is used to identify and match table's lines and plot's items (curves, etc.). Table's row id can
        # not be used because it changes when previous rows are deleted.
        self.current_variable_id = 0

        self.__initialize_variable_table__()
        self.__initialize_plot__()
        self.__initialize_dataset_combobox__()

        self.main_ui.explo_combobox_dataset.currentTextChanged.connect(self.__update_current_dataset__)
        self.main_ui.explo_pushbutton_add_row.clicked.connect(self.__add_new_row_in_variable_table__)
        self.main_ui.explo_tablewidget_variables.cellClicked.connect(self.__change_color__)

    def __initialize_dataset_combobox__(self):
        """Populate the "datasets" combobox with the existing dataset directory names."""
        data_root_dir = self.config.read("DATA_SOURCE", "absolute_root_dir")
        dataset_dirs = self.explo_prvd.explore_root_directory(data_root_dir)
        dataset_dirs.sort(reverse=True)
        for directory in dataset_dirs:
            self.main_ui.explo_combobox_dataset.addItem(directory)

        self.__update_current_dataset__(dataset_dirs[0])

    def __update_current_dataset__(self, dataset_dir: str, ask_confirmation: bool = True):
        self.main_ui.explo_tablewidget_variables.setRowCount(0)
        self.__initialize_plot__()

        # Change current dataset and add a new row
        self.current_dataset = self.explo_prvd.datasets[dataset_dir]
        self.__add_new_row_in_variable_table__()

    ####################################################################################################################
    # "Variables" table

    def __initialize_variable_table__(self):
        """Initialize the table containing the to-be-displayed variables"""
        
        # Set column widths
        self.main_ui.explo_tablewidget_variables.setColumnWidth(self.INSTRUMENT_COL, 130)
        self.main_ui.explo_tablewidget_variables.setColumnWidth(self.VARIABLE_COL, 110)
        self.main_ui.explo_tablewidget_variables.setColumnWidth(self.COLOR_COL, 55)
        self.main_ui.explo_tablewidget_variables.setColumnWidth(self.OFFSET_COL, 90)
        self.main_ui.explo_tablewidget_variables.setColumnWidth(self.MULT_COL, 90)
        self.main_ui.explo_tablewidget_variables.setColumnWidth(self.TIMESHIFT_COL, 95)
        self.main_ui.explo_tablewidget_variables.setColumnWidth(self.VISIBLE_COL, 70)
        self.main_ui.explo_tablewidget_variables.setColumnWidth(self.XORIG_COL, 80)
        self.main_ui.explo_tablewidget_variables.setColumnWidth(self.YORIG_COL, 70)

        self.__add_new_row_in_variable_table__()

    def __add_new_row_in_variable_table__(self):
        """Add a new row in the calibration table. Initialize the cell's content and properties."""
        table = self.main_ui.explo_tablewidget_variables
        row_id = table.rowCount()
        table.insertRow(row_id)

        # Instruments
        instrument_item = QComboBox()
        table.setCellWidget(row_id, self.INSTRUMENT_COL, instrument_item)
        table.cellWidget(row_id, self.INSTRUMENT_COL).currentTextChanged.connect(
            lambda text, row_id=row_id: self.__update_variables_combobox__(combobox_text=text, row=row_id))

        # Variables
        variable_item = QComboBox()
        table.setCellWidget(row_id, self.VARIABLE_COL, variable_item)
        table.cellWidget(row_id, self.VARIABLE_COL).currentTextChanged.connect(
            lambda text, row_id=row_id: self.__apply_variable_change__(row_id=row_id, variable_name=text))

        # Color
        color_item = QTableWidgetItem()
        color_item.setBackground(self.DEFAULT_COLORS[row_id % len(self.DEFAULT_COLORS)])
        table.setItem(row_id, self.COLOR_COL, color_item)

        # Offset
        offset_item = QDoubleSpinBox()
        offset_item.setMinimum(-1000.0)
        table.setCellWidget(row_id, self.OFFSET_COL, offset_item)
        table.cellWidget(row_id, self.OFFSET_COL).valueChanged.connect(
            lambda value, row_id=row_id: self.__apply_variable_change__(row_id=row_id))

        # Multiplicative factor
        mult_item = QDoubleSpinBox()
        mult_item.setValue(1.0)
        mult_item.setMinimum(-1000.0)
        table.setCellWidget(row_id, self.MULT_COL, mult_item)
        table.cellWidget(row_id, self.MULT_COL).valueChanged.connect(
            lambda value, row_id=row_id: self.__apply_variable_change__(row_id=row_id))

        # Time shift
        timeshift_item = QDoubleSpinBox()
        timeshift_item.setMinimum(-1000.0)
        table.setCellWidget(row_id, self.TIMESHIFT_COL, timeshift_item)
        table.cellWidget(row_id, self.TIMESHIFT_COL).valueChanged.connect(
            lambda value, row_id=row_id: self.__apply_variable_change__(row_id=row_id))

        # Visible
        visible_item = QCheckBox()
        visible_item.setChecked(True)
        table.setCellWidget(row_id, self.VISIBLE_COL, visible_item)
        table.cellWidget(row_id, self.VISIBLE_COL).stateChanged.connect(
            lambda state, row_id=row_id: self.__apply_variable_change__(row_id=row_id))

        # X original
        xorig_item = QTableWidgetItem()
        xorig_item.setFlags(Qt.ItemIsEnabled)  # Read only
        table.setItem(row_id, self.XORIG_COL, xorig_item)

        # Y original
        yorig_item = QTableWidgetItem()
        yorig_item.setFlags(Qt.ItemIsEnabled)  # Read only
        table.setItem(row_id, self.YORIG_COL, yorig_item)

        self.__update_instruments_combobox__(row_id)

    def __update_instruments_combobox__(self, row_id: int):
        instrument_combobox = self.main_ui.explo_tablewidget_variables.cellWidget(row_id, self.INSTRUMENT_COL)
        instrument_combobox.clear()

        dataset_dir = self.main_ui.explo_combobox_dataset.currentText()
        if dataset_dir == "":
            return

        dataset = self.explo_prvd.datasets[dataset_dir]
        instrument_logs = dataset.instlogs
        instrument_logs_names = [log_name for log_name in instrument_logs]
        instrument_logs_names.sort()
        for instrument_name in instrument_logs_names:
            instrument_combobox.addItem(instrument_name)

    def __update_variables_combobox__(self, combobox_text: str, row: int):
        variables_combobox = self.main_ui.explo_tablewidget_variables.cellWidget(row, self.VARIABLE_COL)
        variables_combobox.clear()
        instrument_log_name = combobox_text
        instrument_log = self.current_dataset.instlogs[instrument_log_name]
        variable_names = instrument_log.get_variables()
        for variable_name in variable_names:
            variables_combobox.addItem(variable_name)

    def __change_color__(self, row: int, column: int):
        if column != self.COLOR_COL:
            return

        color = QColorDialog.getColor()
        self.main_ui.explo_tablewidget_variables.item(row, self.COLOR_COL).setBackground(color)
        self.__apply_variable_change__(row)

    def __get_row_dataframe__(self, row_id: int) -> pd.DataFrame:
        table = self.main_ui.explo_tablewidget_variables

        # Get instrument log
        instrument_log_name = table.cellWidget(row_id, self.INSTRUMENT_COL).currentText()
        instrument_log = self.current_dataset.instlogs[instrument_log_name]

        # Get variable name
        variable_name = table.cellWidget(row_id, self.VARIABLE_COL).currentText()

        timeseries = instrument_log.get_timeseries(variable_name)

        return timeseries

    def __apply_variable_change__(self, row_id: int, variable_name: str = None) -> None:
        """Read the information related to the row_id and call plot function

        Parameters
        ----------
        row_id
        variable_name
        """
        # The variable combobox is cleared before variables of the newly-selected instrument are written.
        # This function is called on variable-combobox edition, thus it is also called on combobox clear.
        # So in this case, do nothing, wait for the call related to the filling of the combobox
        if variable_name == "":
            return

        table = self.main_ui.explo_tablewidget_variables

        timeseries = self.__get_row_dataframe__(row_id)

        # Get color
        color = table.item(row_id, self.COLOR_COL).background().color()

        # Get position adjustment variables (offset, multiplicative factor and time shift)
        offset = table.cellWidget(row_id, self.OFFSET_COL).value()
        mult = table.cellWidget(row_id, self.MULT_COL).value()
        timeshift = table.cellWidget(row_id, self.TIMESHIFT_COL).value()

        # Get variable visibility
        visible = table.cellWidget(row_id, self.VISIBLE_COL).isChecked()

        try:
            self.__update_plot__(timeseries, row_id, color, offset, mult, timeshift, visible)
        except TypeError:
            self.main_ui.statusbar.showMessage("Failed to plot [" + variable_name + "]", msecs=3000)

    def __update_xy_original__(self, instant: datetime.datetime):
        table = self.main_ui.explo_tablewidget_variables
        for row_id in range(table.rowCount()):
            # Get X orig (original non-shifted datetime value)
            timeshift_sec = table.cellWidget(row_id, self.TIMESHIFT_COL).value()
            x_orig = instant - datetime.timedelta(seconds=timeshift_sec)
            table.item(row_id, self.XORIG_COL).setText(x_orig.strftime("%H:%M:%S.%f")[:-5])

            # Get Y orig (original non-shifted variable value)
            df = self.__get_row_dataframe__(row_id).copy()
            df = df[df["datetime"] <= instant]
            if len(df.index) == 0:
                y_orig = "Out of range"
            else:
                y_orig = df.iloc[-1]["value"]
            if isinstance(y_orig, float):
                y_orig = "{:.4f}".format(y_orig)
            table.item(row_id, self.YORIG_COL).setText(y_orig)

    ####################################################################
    # Plot

    def __initialize_plot__(self) -> None:

        self.plot_item = pg.PlotItem(axisItems={'bottom': utils.TimeAxisItem(orientation='bottom')})

        self.step_curves = dict()

        self.main_ui.explo_graphicsview_top.setCentralItem(self.plot_item)

        self.lines_items = []

        # Vertical line following the cursor
        self.cursor_vline = pg.InfiniteLine(angle=90, movable=False, pen=pg.mkPen(style=Qt.DotLine))
        self.plot_item.addItem(self.cursor_vline, ignoreBounds=True)

        self.measure_vline = pg.InfiniteLine(angle=90, movable=False)
        self.plot_item.addItem(self.measure_vline, ignoreBounds=True)

    def __update_plot__(self, timeseries: pd.DataFrame, row_id: int, color: QColor,
                        offset: float, mult: float, timeshift_sec: float,
                        visible: bool) -> None:

        # Get the to-be-modified curve
        if row_id in self.step_curves:
            step_curve = self.step_curves[row_id]
        else:
            step_curve = pg.PlotCurveItem()
            self.step_curves[row_id] = step_curve
            self.plot_item.addItem(step_curve)
            step_curve.scene().sigMouseClicked.connect(lambda event: self.__mouse_clicked__(event))
            step_curve.scene().sigMouseMoved.connect(lambda event: self.__mouse_moved__(event))

        # Set curve visibility
        if not visible:
            step_curve.hide()
        else:
            step_curve.show()

        # --- X values ---
        # As it is a _step_ curve, add a last datetime point to determine the end of the last step. This is the datatime
        # of the last available data of the dataset, plus one second.
        last_datetime = self.current_dataset.last_data_datetime + datetime.timedelta(seconds=1)
        x_values = timeseries['datetime'].copy()
        x_values = x_values.append(pd.Series([last_datetime]))

        # Apply time shift
        x_values = x_values + datetime.timedelta(seconds=timeshift_sec)

        # --- Y values ---
        # Get original value if it is a numeric, otherwise get its coded integer version.
        if is_numeric_dtype(timeseries["value"]):
            y_values = list(timeseries["value"])
        else:
            y_values = list(timeseries["value_int"])

        # Apply multiplicative factor
        y_values = [y * mult for y in y_values]

        # Apply Y-axis offset
        y_values = [y + offset for y in y_values]

        # Set data to the plot
        step_curve.setData(x=utils.pd_time_to_epoch_ms(x_values),
                           y=y_values,
                           pen=color,
                           stepMode=True)

    def __mouse_clicked__(self, event: MouseClickEvent) -> None:
        """Function triggered when the user clicks on the plot. Display a vertical line under the mouse click and call
        function ``__update_xy_original__``

        Parameters
        ----------
        event: pyqtgraph.MouseClickEvent
        """
        pos = event.scenePos()
        mouse_point = self.step_curves[0].getViewBox().mapSceneToView(pos)
        instant = datetime.datetime.fromtimestamp(mouse_point.x(), tz=datetime.timezone.utc)
        self.measure_vline.setPos(mouse_point.x())
        self.__update_xy_original__(instant)

    def __mouse_moved__(self, pos: QPointF) -> None:
        """Function triggered when the user's mouse cursor hovers over the plot. Display a vertical line where the
        cursor is.

        Parameters
        ----------
        pos: PyQt5.QtCore.QPointF
        """
        # if self.step_curves[0].sceneBoundingRect().contains(pos):
        mouse_point = self.step_curves[0].getViewBox().mapSceneToView(pos)
        self.cursor_vline.setPos(mouse_point.x())